
import gwinc_kagra.kagra_sensitivity as kagra_sensitivity

import numpy as np
from numpy import pi,sqrt

def sensitivity(conf, verbose = False):
    """Calculate sensitivity with the original pythob code 
        with the specified configuration
    """

    phidet = pi/180*86.5  # detune phase
    xi     = pi/180*135.1 # homodyne phase
    rSRM   = 0.92         # amplitude reflectivity of SRM
    Tm     = 22           # temperature at TM [k]
    I0attn = 1            # attenuated from maximum power
    len3   = 35e-2        # fiber length
    wsafe  = 12.5765      # safety factor for fiber radius
    mass   = 22.8         # mirror mass
    SQZ    = 0
    FC     = 0
    LF     = 0

    #wsafe = 12.5765*4*(r2/sqrt(mass/22.8)/1.6e-3)**2

    if   conf == 'DRSE':        #detuned RSE
    	phidet = pi/180*(90-3.5)
        xi     = pi/180*135.1

    elif conf == 'BRSEhomo':    #broadband RSE with homodyne
    	phidet = pi/180*90
     	xi     = pi/180*119.1

    elif conf == 'BRSE':        #broadband RSE without homodyne
      	phidet = pi/180*90
      	xi     = pi/180*90

    elif conf == 'BRSEsqz':     #broadband RSE with FD squeezing
      	phidet = pi/180*90
      	xi     = pi/180*90
        SQZ    = 16 #6
        FC     = 45 #100

    elif conf == 'LF':          #KAGRA+ LF
      	phidet = pi/180*(90-28.5)
      	xi     = pi/180*133.6
        LF     = 1
        Tm     = 23.6
        I0attn = 0.928
        rSRM   = sqrt(0.955)
        len3   = 99.8e-2
        wsafe  = 0.995

    elif conf == 'HF':          #KAGRA+ HF
      	phidet = pi/180*(90-0.1)
      	xi     = pi/180*97.1
        Tm     = 20.8
        rSRM   = sqrt(0.907)
        len3   = 20.1e-2
        wsafe  = 30.25
        SQZ    = 10

    elif conf == '40kg':        #KAGRA+ 40kg
      	phidet = pi/180*(90-3.5)
      	xi     = pi/180*123.2
        Tm     = 21.0
        rSRM   = sqrt(0.922)
        len3   = 28.6e-2
        mass   = 40
        wsafe  = 13.45

    elif conf == 'FDSQZ':       #KAGRA+ FDSQZ
      	phidet = pi/180*(90-0.2)
      	xi     = pi/180*93.1
        Tm     = 21.3
        rSRM   = sqrt(0.832)
        len3   = 23.0e-2
        wsafe  = 17.8
        SQZ    = 10
        FC     = 30

    else:
        print 'Unknown configuration: ',conf,' DRSE is assumed'
    
    nn = kagra_sensitivity.sensitivity(phidet,xi,rSRM,Tm,I0attn,\
                                       len3,wsafe,mass,LF=LF,SQZ=SQZ,FC=FC)

    if verbose:
        print
        print 'kagra_sensitivity:'
        print 'Power at BS:       %.0f W'  %  nn['I0']
        print 'Fiber diameter:    %.2f mm' % (nn['r2']*2000)

        if FC:
            detune(nn, FC, rSRM, mass)

    return nn


def detune(nn, FC, rSRM, mass):
    c        = 2.99792458e8             # speed of light (m/s)
    L        = 3e3                      # arm length (m)
    lamb     = 1.064e-6                 # laser wave length (m)
    I0       = nn['I0']
    trans    = 0.004                    # power transmittivity of ITM
    lambsq2  = 30e-6                    # inside the filter cavity
    Lfilter  = FC
    gamma    = trans*c/(4*L)            # cavity pole (Hz)
    gammaRSE = (1+rSRM)/(1-rSRM)*gamma	# RSE linewidth

    # effective input power at BS for RSE (W)
    I0RSE    = (1+rSRM)/(1-rSRM)*I0
    # SQL reaching frequency for BRSE (rad/s)
    omegaSQL = sqrt((-gammaRSE**2+sqrt(gammaRSE**4+64*pi*c*I0RSE \
                                             /(mass*lamb*L**2)))/2)
    # loss parameter
    epsilonfilter = 4/(2+sqrt(2+2*sqrt(1+(4*omegaSQL*Lfilter/(c*lambsq2))**4)))

    # line width of filter cavity (rad/s)
    kappafilter   = sqrt(2/(2-epsilonfilter) \
                   /sqrt(1-epsilonfilter))*omegaSQL/sqrt(2)

    # detuning of filter cavity (Hz)
    detune        = sqrt(1-epsilonfilter) *kappafilter

   #sqmax = (np.max(20*np.log10(hsh/hshsq)))
   #sqhf  = (20*log10(hsh[-5]/hshsq[-5]))

    print 'FC half bandwidth: %.2f Hz' % (kappafilter/(2*pi))
    print 'FC detuning:       %.2f Hz' % (detune/(2*pi))

def noise_gwinc(n0, nn):
    """Get the noise component with the name in the original python code
    """

    if n0 == 'hseis':
        return sqrt(nn['Seismic']+nn['Newtonian Gravity'])

    if n0 == 'hsusp':
        return sqrt(nn['Suspension Thermal'])

    if n0 == 'hmir':
        return sqrt(nn['Coating Brownian']+\
                    nn['Substrate Brownian']+nn['Substrate Thermo-Elastic'])

    if n0 == 'hsh':
        return sqrt(nn['Quantum Vacuum'])

    if n0 == 'htotal':
        return sqrt(nn['Total'])

    return nn['Freq']

