
import scipy
import numpy as np
from numpy import pi

def susptherm(f, ifo):
    """Suspention thermal noise.
    """
    #####suspension thermal noise#####
    #see VIR-015E-09 (2009)
    #violin mode parameters
    #
    # Parameters in kagra_sensitivity.py
    #
    #g=9.80665             #gravitational acceleration (m/s**2)
    #mass=22.8
    #wsafe=12.5765
    #LF = False
    #Tm=22
    #
    #####suspension parameters
    #
    #rhos=4.0e3       #density of suspension wire (kg/m**3) #sapphire
    #
    #mass (kg)
    #m1=20.5          #IM
    #if LF:
    #    m1=20.5*4		# for KN KAGRA-LF
    #m2=0.055*4       #blade spring
    #m3=mass          #TM
    #
    #resonant frequency of blade spring (Hz)
    #wh2=2*pi*2e3     #horizontal
    #fmg=14.5         #vertical
    #
    #radius of wire (m)
    #r1=0.6e-3/2     #wire1 between IM and MN; scale with IM mass
    #r2=1.6e-3/2*(mass/22.8)**(1./2)*(wsafe/12.5765)**(1./2)
    #wire2 between blade spring and TM; scale with mirror mass
    # (wsafe for safety factor; default value calculated with modulus of 
    #  rupture to be 350 MPa (360e6*pi*r2**2/g*4/22.8))
    #if LF:
    #    r1=0.2e-3/2 # for KN KAGRA-LF
    #   #r2=3.2e-4/2 # for KN KAGRA-LF
    #
    #Youngs modulus (Pa)
    #Y1=1.3e11        #wire1 #CuBe
    #Y2=4.0e11        #blade spring #sapphire
    #Y3=4.0e11        #wire2 #sapphire
    #
    #length of wire (m)
    #l1=26.1e-2       #wire1
    #if LF:
    #	l1=26.1e-2*3       #wire1 for KN KAGRA-LF
    #l2=0.5e-2        #blade spring
    #l3=35e-2         #wire2
    #
    #loss angle of wire
    #phi1=5e-6        #wire1
    #phi2=7e-7        #blade spring
    #phi3=2e-7        #wire2
    # 
    #temperature [k]
    #T1=16            #wire1
    #T2=16            #blade spring
    #T3=(T2+Tm)/2     #wire2
    #
    #loss angle normalized by T3
    #phih1=phi1*T1/T3 #horizontal wire1
    #phih2=phi2*T2/T3 #horizontal blade spring
    #phih3=phi3       #horizontal wire2
    #phiv1=phi1*T1/T3 #vertical wire1
    #phiv2=phi2*T2/T3 #vertical blade spring
    #phiv3=phi3       #vertical wire2
    #
    #coupling from vertical mode to horizontal mode
    #VHC=1./200
    #
    #number of wires
    #N1=4
    #N2=4
    #N3=4
    #
    #number of bending points
    #b1=2
    #b2=2
    #b3=2
    #

    # Parameters in GWINC format
    kb = scipy.constants.k  # Boltzmann constant (J/K)
    g  = scipy.constants.g  # gravitational acceleration (m/s**2)

    # Comment this line out if we use scipy
    kb = 1.38064852e-23     # Boltzmann constant (J/K)

    Tm    = ifo.Materials.Substrate.Temp #temperature at TM [k]
    L     = ifo.Infrastructure.Length       # Length of arm cavities [m]
    omega = 2*pi*f
    fres  = len(f)
    i     = 1j
    #
    #####suspension parameters
    #
    # Note stage numbering: mirror is at beginning of stack, not end
    #
    rhos  = ifo.Suspension.Sapphire.Rho
    mass  = ifo.Suspension.Stage[0].Mass
    #
    #resonant frequency of blade spring (Hz)
    wh2   = ifo.Suspension.BladeSpring.wh2*2*pi  #horizontal
    fmg   = ifo.Suspension.BladeSpring.fmg       #vertical
    #
    #mass (kg)
    m1    = ifo.Suspension.Stage[2].Mass       #IM
    m2    = ifo.Suspension.Stage[1].Mass       #blade spring
    m3    = ifo.Suspension.Stage[0].Mass       #TM
    #
    #radius of wire (m)
    r1    = ifo.Suspension.Stage[2].WireRadius #wire1: between IM and MN
    r2    = ifo.Suspension.Stage[0].WireRadius #wire2: blade spring and TM
    #
    #Youngs modulus (Pa)
    Y1    = ifo.Suspension.CuBe.Y              #wire1 #CuBe
    Y2    = ifo.Suspension.Sapphire.Y          #blade spring #sapphire
    Y3    = ifo.Suspension.Sapphire.Y          #wire2 #sapphire
    #
    #length of wire (m)
    l1    = ifo.Suspension.Stage[2].Length     #wire1
    l2    = ifo.Suspension.Stage[1].Length     #blade spring
    l3    = ifo.Suspension.Stage[0].Length     #wire2
    #
    #loss angle of wire
    phi1  = ifo.Suspension.Stage[2].Phi        #wire1
    phi2  = ifo.Suspension.Stage[1].Phi        #blade spring
    phi3  = ifo.Suspension.Stage[0].Phi        #wire2
    #
    #temperature [k]
    T1    = ifo.Suspension.Temp                #wire1
    T2    = ifo.Suspension.Temp                #blade spring
    T3    = (T2+Tm)/2                          #wire2
    #
    #loss angle normalized by T3
    phih1 = phi1*T1/T3  #horizontal wire1
    phih2 = phi2*T2/T3  #horizontal blade spring
    phih3 = phi3        #horizontal wire2
    phiv1 = phi1*T1/T3  #vertical wire1
    phiv2 = phi2*T2/T3  #vertical blade spring
    phiv3 = phi3        #vertical wire2
    #
    #coupling from vertical mode to horizontal mode
    VHC   = ifo.Suspension.VHCoupling
    #
    #number of wires
    N1    = ifo.Suspension.Stage[2].NWires
    N2    = ifo.Suspension.Stage[1].NWires
    N3    = ifo.Suspension.Stage[0].NWires
    #
    #number of bending points
    b1    = ifo.Suspension.Stage[2].BendingPoints
    b2    = ifo.Suspension.Stage[2].BendingPoints
    b3    = ifo.Suspension.Stage[2].BendingPoints


    lsus=l3
    Ewire=Y3*(1+i*phi3)    #complex Youngs modulus (Pa)
    Icwire=pi*r2**4/4      #area moment of inertia (m**2)
    tension=mass*g/N3      #tension per wire (N)
    rholine=rhos*pi*r2**2  #linear density (kg/m)
    
    # wave number of elastic fibers (1/m)
    ks=np.sqrt((-tension+np.sqrt(tension**2+4*Ewire*Icwire*rholine*omega**2)) \
               /(2*Ewire*Icwire))
    
    # wave number associated with the flexural stiffness
    ke=np.sqrt((tension+np.sqrt(tension**2+4*Ewire*Icwire*rholine*omega**2)) \
               /(2*Ewire*Icwire))
    
    # effective spring constant (N/m)
    # Calculated by K. Somiya and confirmed by K. Komori.
    # The spring constant Eq.(48) in PPPnote (VIR-015E-09) is 
    # the limitting value of ke*lsus -> infinity in our case.
    Ksusp=N3*Ewire*Icwire*ke*ks*(ke**2+ks**2)  \
        *(ke*np.cos(ks*lsus)*np.sinh(ke*lsus)+ \
          ks*np.sin(ks*lsus)*np.cosh(ke*lsus)) \
          /(2*ke*   ks* (1-np.cos(ks*lsus)*np.cosh(ke*lsus)) \
            +(ke**2-ks**2)*np.sin(ks*lsus)*np.sinh(ke*lsus))
    
    #spring constant and resonant frequency
    Ecube=Y1*(1+i*phih1)           #complex Youngs modulus (Pa)
    I1=pi*r1**4/4                  #area moment of inertia (m**2)
    tensioncube=(m1+m2+m3)*g/N1    #tension per CuBe wire (N)
    
    #hor. spring constatnt of wire1 (N/m)
    kh1=N1*tensioncube/(l1-2*np.sqrt(Ecube*I1/tensioncube))
    kh2=m2*wh2**2*(1+i*phih2)      #hor. spring constatnt of blade spring (N/m)
    kh3=Ksusp
    
    wv1=np.sqrt(pi*r1**2*Y1*N1/(l1*m1))   #ver. wire1 (Hz)
    kv1=m1*wv1**2*(1+i*phiv1)             #ver. spring constatnt of wire1 (N/m)
    
    wv2=2*pi*fmg*np.sqrt((m2+m3)/m2)      #ver. blade spring (Hz)
    
    kv2=(m2+m3)*(2*pi*fmg)**2*(1+i*phiv2) #ver. spring constatnt of
                                          #     blade spring (N/m)
    
    wv3=np.sqrt(pi*r2**2*Y3*N3/(l3*m3))   #ver. wire2 (Hz)
    kv3=m3*wv3**2*(1+i*phiv3)             #ver. spring constatnt of wire2 (N/m)
    
    #matrix of horizontal spring constatnt (N/m)
    Dh11=-m1*omega**2+kh1+kh2
    Dh12=-kh2*np.ones(fres)
    Dh13=0*np.ones(fres)
    Dh21=Dh12
    Dh22=-m2*omega**2+kh2+kh3
    Dh23=-kh3
    Dh31=Dh13
    Dh32=Dh23
    Dh33=-m3*omega**2+kh3
    
    # matrix of vertical spring constatnt (N/m)
    Dv11=-m1*omega**2+kv1+kv2
    Dv12=-kv2*np.ones(fres)
    Dv13=0*np.ones(fres)
    Dv21=Dv12
    Dv22=-m2*omega**2+kv2+kv3
    Dv23=-kv3*np.ones(fres)
    Dv31=Dv13
    Dv32=Dv23
    Dv33=-m3*omega**2+kv3
    
    # for-loop omitted from matlab version for speed up (no change in results)
    # matrix of horizontal impedance (N/m/Hz) and inversion
    Zh=np.array([[Dh11,Dh12,Dh13],[Dh21,Dh22,Dh23],[Dh31,Dh32,Dh33]])/(i*omega)

    # gives invZh[:,:,k]=np.linalg.inv(Zh[:,:,k])
    invZh=np.transpose(np.linalg.inv(np.transpose(Zh)))

    #matrix of vertical impedance (N/m/Hz) and inversion
    Zv=np.array([[Dv11,Dv12,Dv13],[Dv21,Dv22,Dv23],[Dv31,Dv32,Dv33]])/(i*omega)
    invZv=np.transpose(np.linalg.inv(np.transpose(Zv)))

    ###calculation of eq.(63) in PPPnote (VIR-015E-09)###
    # The fluctuation of TM is derived from 3*3 component of 
    #  the inverse matrix of impedance.
    Fsqh=4*kb*T3*np.real(invZh)/omega**2
    Sthh=Fsqh[2,2,:]

    Fsqv=4*kb*T3*np.real(invZv)/omega**2
    Sthv=Fsqv[2,2,:]
    
    #strain of thermal noise with 4 TMs
    Ssusph = np.abs(Sthh)
    Ssuspv = np.abs(Sthv)

    # hsusp value slightly different from matlab code at 1e-11 level

    # 4 masses, turn into gravitational wave strain
    noise = 4 * (Ssusph+Ssuspv*VHC*VHC) * ifo.gwinc.dhdl_sqr

    return noise
